/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.
   
   DEU.H - Main game defines.
   */

/* the includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <conio.h>
#include <ctype.h>
#ifndef _MSC_VER
#include <unistd.h>
#include <bios.h>
#endif
#include <grx20.h>
#include <bgiext.h>

#include "deu-go32.h"

#if defined(__WATCOMC__) || defined(_MSC_VER)
#define M_PI 3.14159265359
#endif

#define DEU_VERSION		"5.21"
#define DETH_VERSION	"4.17"

typedef short int			BCINT;
typedef unsigned short int	UBCINT;
typedef unsigned char		BYTE;


/* define some new colors */
enum
{
#ifndef BLACK
	BLACK,
	BLUE,
	GREEN,
	CYAN,
	RED,
	MAGENTA,
	BROWN,
	LIGHTGRAY,
	DARKGRAY,
	LIGHTBLUE,
	LIGHTGREEN,
	LIGHTCYAN,
	LIGHTRED,
	LIGHTMAGENTA,
	YELLOW,
	WHITE,
#endif
	DARKBLUE = 16,
	DARKGREEN,
	DARKRED,
	DARKMAGENTA,
	GRAY,
	DARKERGRAY,
	ORANGE,
	SECTORTAGGED,
	SECTORSECRET,
	SECTORPAINFUL,
	SECTORLIGHT,
	LINEDEFTAGGED,
	LINEDEFSECRET,
	LINEDEFNOSOUND,
	LINEDEFNOPASS,
	LINEDEFNOMAP,
	GFXBLUE,
	GFXWADLIST,
	THINGPATH,
	THINGTOPATH,
	GRIDCOLOR,
	FLOORGRIDCOLOR,

	PALETTE_SIZE
};

#define MENUARRAYMAX	40

#include "struct_h.h"

/* constants for filtering out things based on difficulty */

#define TF_EASY			1
#define TF_MEDIUM		2
#define TF_HARD			4
#define TF_FIGHTER		32
#define TF_CLERIC   	64
#define TF_MAGE   		128
#define TF_SINGLE   	256
#define TF_COOP   		512
#define TF_DEATHM   	1024

/* constants for telling what we have to Undo */
#define BTHINGS        	1
#define BLINEDEFS      	2
#define BSIDEDEFS      	4
#define BVERTEXES      	8
#define BSECTORS       	16


/*
   the directory structure is the structure used by DOOM to order the
   data in it's WAD files
   */

typedef struct Directory huge *DirPtr;
struct Directory
{
    long start;			/* offset to start of data */
    long size;			/* byte size of data */
    char name[ 8];		/* name of data block */
};



/*
   The wad file pointer structure is used for holding the information
   on the wad files in a linked list.
   
   The first wad file is the main wad file. The rest are patches.
   */

typedef struct WadFileInfo huge *WadPtr;
struct WadFileInfo
{
    WadPtr next;		/* next file in linked list */
    char *filename;		/* name of the wad file */
    FILE *fileinfo;		/* C file stream information */
    char type[ 4];		/* type of wad file (IWAD or PWAD) */
    long dirsize;		/* directory size of WAD */
    long dirstart;		/* offset to start of directory */
    DirPtr directory;	/* array of directory information */
};



/*
   the master directory structure is used to build a complete directory
   of all the data blocks from all the various wad files
   */

typedef struct MasterDirectory huge *MDirPtr;
struct MasterDirectory
{
    MDirPtr next;			/* next in list */
    WadPtr wadfile;			/* file of origin */
    struct Directory dir;	/* directory data */
};



/*
   the selection list is used when more than one object is selected
   */

typedef struct SelectionList *SelPtr;
struct SelectionList
{
    SelPtr next;		/* next in list */
    BCINT objnum;		/* object number */
};


/*
   syntactic sugar
   */
typedef BCINT Bool;		/* Boolean data: true or false */


/*
   description of the command line arguments and config file keywords
   */

typedef struct
{
    char *short_name;		/* abbreviated command line argument */
    char *long_name;		/* command line arg. or keyword */
    enum {                  /* type of this option */
	OPT_BOOLEAN,			/* boolean (toggle) */
	OPT_INTEGER,			/* integer number */
	OPT_STRING,				/* character string */
	OPT_STRINGACC,			/* character string, but store in a list */
	OPT_STRINGLIST,         /* list of character strings */
	OPT_END                 /* end of the options description */
    } opt_type;                    
    char *msg_if_true;		/* message printed if option is true */
    char *msg_if_false;		/* message printed if option is false */
    void *data_ptr;         /* pointer to the data */
} OptDesc;


/* generic string list type  -- used for holding level names,
   ftexture sections, texture sections */
	typedef struct _SList {
	char *string;
	struct _SList *next;
} *SList;

/* sector type type (!) (actually a list type) */
   typedef struct _sector_type {
	char *shortname, *longname;
	BCINT type;
	struct _sector_type *next;
} sector_type;

typedef struct _sector_class {
	char *name;
	sector_type *types;
	struct _sector_class *next;
} sector_class;

/* type for a list of linedef types */
typedef struct _ld_type {
	char *shortname, *longname;
	char *argtext[5], *enums[5];
	char args[6];
	BCINT type;
	struct _ld_type *next;
} ld_type;

/* type for the list of linedef type classes (phew!) */
typedef struct _ld_class {
	char *name;
	ld_type *types;
	struct _ld_class *next;
} ld_class;

/* as above, but for a thing */
typedef struct _thing_type {
    BCINT	type, colour, radius, extra;
	BCINT	pathtype;
	char	args[6];
	char	*argtext[5], *enums[5];
    char	*sprite;
    char 	*name;
    struct _thing_type *next;
} thing_type;

enum
{
	TX_POLYSPOT		= 0x001,		// Thing uses angle as id
	TX_DEATHSCRIPT	= 0x002,		// Thing executes a script on death instead of a special
	TX_NOANGLE		= 0x004,		// Don't show the thing's angle because it doesn't use it
	TX_START		= 0x008,		// Thing is a single-player/coop start
	TX_DMSTART		= 0x010,		// Thing is a deathmatch start
	TX_SECTORACTION	= 0x020,		// Thing is a sector action
	TX_RESIZER		= 0x040,		// Thing gets its radius from its first argument
	TX_PATH			= 0x080,		// Thing's first tid is the next node in a path
	TX_TOPATH		= 0x100,		// Thing's first tid points into a path, but it is not on a path
	TX_NOPIC		= 0x200,		// Prefer not to draw thing's picture
};

/* and a list of thing classes */
typedef struct _thing_class {
	char *name;
	thing_type *types;
	struct _thing_class *next;
} thing_class;

/* as above, but for a spawnable thing */
typedef struct _spawn_type {
	BCINT	type;
	char	*sprite;
	char 	*name;
	struct _spawn_type *next;
} spawn_type;

/* and a list of spawnable thing classes */
typedef struct _spawn_class {
	char *name;
	spawn_type *types;
	struct _spawn_class *next;
} spawn_class;

/* as above, but for a scheme */
typedef struct _scheme_type {
    BCINT	type;
    char 	*name;
    char	*normal;
    char	*upper;
    char	*lower;
    char	*floor;
    char	*ceiling;
    struct _scheme_type *next;
} scheme_type;

/* and a list of scheme classes */
typedef struct _scheme_class {
	char *name;
	scheme_type *types;
	struct _scheme_class *next;
} scheme_class;

/* list of key types */
typedef struct _key_type
{
	BCINT	type;
	BCINT	mask;
	char	*name;
	char	*shortOn, *shortOff;
	struct _key_type *next;
} key_type;

/* used for binary searching of key types */
typedef struct
{
	BCINT type;
	key_type *key;
} key_ptr;


/* list of user-defined enumerations */
typedef struct _enum_type
{
	char		*name;
	key_type	*keys;
	key_ptr		*array;
	int			count;
	struct _enum_type	*next;
} enum_type;

/* prefab settings for teleport source and destination */
typedef struct
{
	char Texture[9];
	char Side[9];
	char InsetSide[9];
	BCINT Decoration;
	BCINT Top;
	BCINT Bottom;
	BCINT Light;
	BCINT InsetDepth;
	BCINT InsetRadius;
	BCINT InsetPoints;
} TeleportPrefab;

/*
   the macros and constants
   */

/* name of the configuration file */
/* name of the log file (debug mode) */
#define DEU_CONFIG_FILE		"ZETH.INI"
#define DEFAULT_CONFIG_FILE	"ZETHDEF.INI"
#define DEU_LOG_FILE		"ZETH.LOG"


/* convert screen coordinates to map coordinates */
#define MAPX(x)			(OrigX + (((x) - (int)ScrCenterX) / Scale))
#define MAPY(y)			(OrigY + (((int)ScrCenterY - (y)) / Scale))

/* convert map coordinates to screen coordinates */
#define SCREENX(x)		(ScrCenterX + (int) (((x) - OrigX) * Scale))
#define SCREENY(y)		(ScrCenterY + (int) ((OrigY - (y)) * Scale))

#define MAX(a, b)       (((a) > (b)) ? (a) : (b))
#define MIN(a, b)       (((a) < (b)) ? (a) : (b))

/* object types */
#define OBJ_THINGS		1
#define OBJ_LINEDEFS	2
#define OBJ_SIDEDEFS	3
#define OBJ_VERTEXES	4
#define OBJ_SEGS		5
#define OBJ_SSECTORS	6
#define OBJ_NODES		7
#define OBJ_SECTORS		8
#define OBJ_REJECT		9
#define OBJ_BLOCKMAP	10

/* boolean constants */
#ifndef TRUE
#define TRUE			1
#define FALSE			0
#endif

/* half the size of an object (Thing or Vertex) in map coords */
#define OBJSIZE			7


/* the interfile global variables */

/* from deu.c */
extern FILE *logst;				/* filepointer to the error log */
extern SList LevelNameFormat;	/* format of level name structure */
extern BCINT InitialScale;		/* initial zoom factor for map */
extern BCINT TempCloseToLine;   /* temporary variable */	
extern BCINT CloseToLine;		/* how close to a line does a vertex have to be to merge with line */	
extern BCINT VideoMode;	  		/* default video mode for VESA cards */
extern BCINT RealVideoMode;	  	/* the actual, chosen default video mode for VESA cards, this is used with graphical front-end */
extern BCINT Increment;			/* increment for changing sectors with hot keys */
extern BCINT InitGrid;			/* current grid setting, used for starting deth as we left it */
extern BCINT GridScale;			/* current grid setting */
extern BCINT BarStyle;			/* button-bar style */
extern BCINT Gamma;				/* gamma level 0-4 */
extern BCINT StretchDelay;		/* how long to hold left mouse down until stretch box started */
extern BCINT Sensitivity;		/* ? */
extern BCINT MoveSpeed;			/* ? */
extern BCINT DefaultLargeScroll;/* amount of pixels to scroll when mouse moves to edge of screen */
extern BCINT MenuDefault;		/* last item used within a menu, used for menus that return i.e thing flags */
extern BCINT NumPatchWads;		/* number of patch wads open */
extern BCINT NumConfigPatchWads;
extern BCINT EdMode;			/* current editing mode - used for starting deth in last mode edited */
extern int NumSingleStarts;		/* number of different singleplayer (coop) start types */
extern Bool ChooseGame;			/* should deth ask which game to edit ? */
extern Bool GraphFront;			/* should we use a graphical style front-end ? */
extern Bool Select0;           	/* select object 0 by default when switching modes */
extern Bool Debug;				/* are we debugging? */
extern Bool SwapButtons;		/* swap right and middle mouse buttons */
extern Bool Quiet;				/* don't play a sound when an object is selected */
extern Bool Quieter;			/* don't play any sound, even when an error occurs */
extern Bool Expert;				/* don't ask for confirmation for some operations */
extern Bool QisQuit;			/* Quit on Key 'Q' or Save on 'Q' */	
extern Bool VertConf;			/* don't ask for confirmation on Vertices merging even if in expert mode */
extern Bool FakeCursor;			/* use a "fake" mouse cursor */
extern Bool AdditiveSelBox;		/* additive selection box or select in box only? */
extern Bool ThingAngle;			/* draw things with an arrow ? */
extern Bool UsePromptOnly;		/* should we start the command at the prompt ? */
extern Bool DispSprite;   		/* display pretty pictures ? */ 
extern Bool AltInfo;   			/* show an alternative information bar ? */ 
extern Bool Registered;			/* registered or shareware WAD file? */
extern Bool DoingBehavior;		/* temp variable */
extern Bool KeepInfo;			/* should we keep the MAPINFO and SNDINFO lumps */
extern Bool CloseOnUse;			/* close the button-bar upon it's use ? */
extern Bool ShowDistance;		/* show the linedefs lengths in vertex mode ? */
extern Bool NestMenus;			/* draw the pulldown menus from the top or center */
extern Bool AutoMap;			/* draw linedef mode in a doom style automap */
extern Bool ShowThings;			/* draw the things with sprites ? */
extern Bool GridDashed;			/* draw the grid with dashed lines ? */
extern Bool SaveIniFile;		/* save the settings to the ini file when you exit deth ? */
extern Bool CopySideDefs;		/* copy sidedef info when copying linedefs ? */
extern Bool CreatingNewLevel;	
extern Bool SpecialEffects;		/* turn off some error checks */
extern Bool DeleteSideLines;	/* can deleting sidedefs possibly delete lines? */
extern Bool DeleteLineVerts;	/* can deleting linedefs possibly delete vertices? */
extern Bool PersistantMenus;	/* editing menus persist until canceled? */
extern char *BackupText;		/* text to show on alternative info-bar */
extern char LevelName[];		/* what level we are editing ? */
extern char *MainWad;			/* name of the main wad file */
extern char *Doom;				/* path to the Doom config file */
extern char *Doom2;				/* path to the Doom2 config file */
extern char *Heretic;			/* path to the Heretic config file */
extern char *Hexen;				/* path to the Hexen config file */
extern char *Strife;			/* path to the Strife config file */
extern char *Game;				/* default game to edit */
extern char *RegTest;			/* level present in registerd version */
extern char *CurrentScheme;		/* current loaded prefab */
extern char *DoomPwads;			/* path to DOOM pwads  */
extern char *Doom2Pwads;		/* path to DOOM2 pwads  */
extern char *HereticPwads;		/* path to HERETIC pwads  */
extern char *StrifePwads;		/* path to STRIFE pwads  */
extern char *HexenPwads;		/* path to HEXEN pwads  */
extern char *Pwads;				/* path to pwads  */
extern char *Cwd;				/* current working directoy */
extern char **PatchWads;		/* list of patch wads to load */
extern char **ConfigPatchWads;	/* list of patch wads to load */
extern char *CfgFile;			/* alternative config file */
extern int GotImages[];
extern unsigned char *WallImages[]; /*JFF change for fix in DEU.C */
extern char *TextEditor;
extern char *ACCDirectory;
extern char *ACCCommand;
extern Bool NonHexenWarning;
extern Bool SmoothMouseScroll;
extern Bool ShowPaths;
extern Bool DetailScroll;
extern Bool PauseBeforeCompile; // 2000/01/29 & 2002/12/08 Ty - optional pause for Notepad

/* from wads.c */
extern WadPtr  WadFileList;		/* list of wad files */
extern MDirPtr MasterDir;		/* the master directory */
extern SList LevelNames;
extern int isalev(char *);
extern int isamark(char *mrk);

/* from edit.c */
extern BCINT LastSectorTag;		/* last sector tag mouse passed over */
extern BCINT LastLineDefTag;	/* last lindef tag mouse passed over */
extern BCINT MinGrid;			/* minimum grid setting */
extern BCINT MaxGrid;			/* maximum grid setting */
extern Bool DraggingObject;		/* are we dragging ? */
extern Bool ShowSprite;			/* do we show graphical info boxes ? */
extern Bool InfoShown;      	/* is the bottom line displayed? */
extern Bool BrowseMap;      	/* should we browse the map? */
extern Bool Caching;      		/* are we cahching texture information ? */
extern int tff;					/* what things to show, based on difficulty level */


/* from gfx.c */
//extern BCINT GfxMode;			/* current graphics mode, or 0 for text */
//extern float Scale;			/* scale to draw map 20 to 1 */
//extern BCINT OrigX;			/* the X origin */
//extern BCINT OrigY;			/* the Y origin */
extern BCINT PointerX;			/* X position of pointer */
extern BCINT PointerY;			/* Y position of pointer */
//extern BCINT ScrMaxX;			/* maximum X screen coord */
//extern BCINT ScrMaxY;			/* maximum Y screen coord */
//extern BCINT ScrCenterX;		/* X coord of screen center */
//extern BCINT ScrCenterY;		/* Y coord of screen center */
extern BCINT GfxMode;			/* current graphics mode, or 0 for text */
extern double Scale;			/* scale to draw map 20 to 1 */
extern double OrigX;			/* the X origin */
extern double OrigY;			/* the Y origin */
extern int ScrMaxX;				/* maximum X screen coord */
extern int ScrMaxY;				/* maximum Y screen coord */
extern int ScrCenterX;			/* X coord of screen center */
extern int ScrCenterY;			/* Y coord of screen center */
extern BYTE DPal[768];			/* Game Palette */

/* from mouse.c */
extern Bool UseMouse;			/* is there a mouse driver? */
Bool MouseClickedArea(BCINT, BCINT, BCINT, BCINT);

/* from things.c */
extern thing_class *Thing_classes;
extern scheme_class *Scheme_classes;
extern spawn_class *Spawn_classes;

/* from names.c */
extern ld_class *Linedef_classes;
extern sector_class *Sector_classes;
extern enum_type *Enum_classes;

/* from levels.c */
extern SList Ftexture_sections;
extern SList Texture_sections;

/*   the function prototypes    */

/* from deu.c */
int main( int, char *[]);
void ParseCommandLineOptions( int, char *[]);
void ParseConfigFileOptions( char *);
void Usage( FILE *);
void Credits( FILE *);
void FunnyMessage( FILE *);
void DEU_Beep( void);
void PlaySound( BCINT , BCINT );
void ProgError( char *, ...);
void LogMessage( char *, ...);
void MainLoop( void);
void MainLoop2( void);
void FrontEnd( void);
void ParseInput( const char *);
void FileRead (const char *, void **, int *);

/* from wads.c */
int isalev(char *);
void OpenMainWad( char *);
void OpenPatchWad( char *);
void CloseWadFiles( void);
void CloseUnusedWadFiles( void);
WadPtr BasicWadOpen( char *);
void BasicWadRead( WadPtr, void huge *, long);
void BasicWadSeek( WadPtr, long);
MDirPtr FindMasterDir( MDirPtr, char *);
MDirPtr AddMapLumps( MDirPtr, char *);
void ListMasterDirectory( FILE *);
void ListFileDirectory( FILE *, WadPtr);
void BuildNewMainWad( char *, Bool);
void WriteBytes( FILE *, void huge *, long);
void CopyBytes( FILE *, FILE *, long);
BCINT  Exists( char *);
void DumpDirectoryEntry( FILE *, char *);
void SaveDirectoryEntry( FILE *, char *);
void SaveEntryToRawFile( FILE *, char *);
void SaveEntryFromRawFile( FILE *, FILE *, char *);

/* from levels.c */
void ReadLevelData();
void ForgetLevelData( void);
void SaveLevelData( char *); 
void ReadWTextureNames( void);
void ForgetWTextureNames( void);
void ReadFTextureNames( void);
void ReadFTextureNamesIn( char *);
void ForgetFTextureNames( void);

/* from wadconv.c */
void InitConversion (void);
void FinishConversion (void);
Bool TranslateLineDef (int linenum, int type, int tag);
void ConvertSectors (void);
int TranslateSectorType (int special);

/* from edit.c */
void MakeLevelName(SList);
void EditLevel(Bool);
void SelectLevel();
void EditorLoop(Bool browsed);
void DrawMap( BCINT , BCINT, Bool );
void GoToObject( BCINT , BCINT );
void FindThing(int, SelPtr *);
void SaveAs(int);
void DEU_CopyFile(const char *, const char *);
void Backup(int);
void Undo();
extern BCINT Count;
extern int last_backup;

/* from icons.c */
void OpenIconFile (const char *name);
int TranslateIcon (int num);
void DrawIcon1 (int x0, int y0, int x1, int y1, const char *picname);
void DrawIconFrac (int x0, int y0, int x1, int y1, const char *picname, long xstep, long ystep);
void RemapIcons (void);

/* from gfx.c */
void InitGfx( void);
Bool SwitchToVGA256( void);
Bool SwitchToVGA16( void);
void TermGfx( void);
void ClearScreen( void);
void ClearMapScreen(BCINT, BCINT);
void SetColor (int);
void SetTextBg (int);
void SetTextCentered (Bool);
void SetWriteMode (int);
void SetLineDotted (Bool);
void SetLineThickness (int);
void DrawMapLine( int , int , int , int );
void DrawMapCircle( BCINT , BCINT , BCINT );
void DrawMapVector( BCINT , BCINT , BCINT , BCINT );
void DrawMapArrow( BCINT , BCINT , UBCINT);
void DrawMapBox( int, int, int, int );
void DrawMapTBox (int, int, int, int);
void DrawMapHollowBox ( int, int, int, int );
void DrawScreenLine( BCINT , BCINT , BCINT , BCINT );
void /*JFF temp*/DrawThickScreenLine( BCINT , BCINT , BCINT , BCINT );
void ScreenBoxArea(int ,int ,int ,int,int );
void ScreenHighlightArea(int ,int ,int ,int, char *,int, int );
void DrawScreenWindow(int ,int ,int ,int, char *, int);
void DrawScreenBox( BCINT , BCINT , BCINT , BCINT );
void DrawScreenBox3D( BCINT , BCINT , BCINT , BCINT );
void DrawScreenBoxHollow( BCINT , BCINT , BCINT , BCINT );
void DrawScreenFrame( BCINT , BCINT , BCINT , BCINT );
void DrawScreenButton( BCINT , BCINT , BCINT , BCINT, char *, Bool );
void DrawScreenButtonIn( BCINT , BCINT , BCINT , BCINT, char *, Bool );
void DrawScreenButtonOut(BCINT, BCINT, BCINT, BCINT, Bool);
void DrawScreenButtonGray(BCINT, BCINT, BCINT, BCINT, char *, Bool);
void DrawScreenMeter( BCINT , BCINT , BCINT , BCINT , float);
void DrawScreenText( BCINT , BCINT , char *, ...);
void DrawScreenTextSmall( BCINT , BCINT , char *, ...);
void DrawScreen3DText( BCINT , BCINT , BCINT, BCINT, char *, ...);
void DrawScreenBar( BCINT , BCINT , BCINT, Bool);
void DrawPointer();
void SetDoomPalette( BCINT );
BCINT TranslateToGameColor( BCINT );
UBCINT ComputeAngle( BCINT , BCINT );
UBCINT ComputeDist( BCINT , BCINT );
void InsertPolygonVertices( BCINT , BCINT , BCINT , BCINT );
//jff change integer io variables to double for better accuracy
void RotateAndScaleCoords( double  *, double  *, double, double);

/* from things.c */
int BCINT_lessp(const void *, const void *);
void index_things(void);
void index_schemes(void);
BCINT GetThingExtra( BCINT );
BCINT GetThingColour( BCINT );
BCINT GetThingPathType (BCINT);
char *GetThingFlag( int, int, int );
char *GetTaggedLineDefFlag( BCINT, BCINT );
char *GetThingName( BCINT );
char *GetSpawnName( BCINT );
char *GetThingSprite( BCINT );
int   GetThingRadius( int, int );
char *GetAngleName( BCINT );
char *GetWhenName( BCINT );
char *GetUnknown3Name( BCINT );
char *GetThingArgs( BCINT type);
char **GetThingArgText( BCINT type);
char **GetThingEnums( BCINT );
void SetThingTID (int thing, int tid);
void index_spawns (void);
void index_keys (void);
BCINT FindFreeTid(void);
int FindThingByTID (int type, int tid);
int SetThingFlagFilter(int);
void InitTIDList (void);
void FixTIDList (void);

/* from names.c */
char *GetObjectTypeName( BCINT );
char *GetEditModeName( BCINT );
char *GetLineDefTypeName( BCINT );
char *GetLineDefArgs( BCINT );
char **GetArgText( BCINT );
char *GetLineDefTypeLongName( BCINT );
char *GetLineDefFlagsName( BCINT );
char *GetThingFlagsName( int, int );
char *GetThingFlagLongName( int, int );
char *GetSectorFlags (BCINT);
char *GetSectorTypeName( BCINT );
char *GetSectorTypeLongName( BCINT );
char *GetHexenAngleName( BCINT );
char *GetNegativeName( BCINT );
char **GetLineDefEnums( BCINT );
int TwiddleFlagsMenu (key_type *, char *, int, int *);
char *GetUserFlagsName( BCINT, char *);
char *GetEnumName( BCINT, char * );
enum_type *FindEnum (char *);
void index_ld_types();

/* from mouse.c */
void CheckMouseDriver( void);
void ShowMousePointer( void);
void HideMousePointer( void);
void GetMouseCoords( BCINT  *, BCINT  *, BCINT  *);
void SetMouseCoords( BCINT , BCINT );
void SetMouseLimits( BCINT , BCINT , BCINT , BCINT );
void ResetMouseLimits( void);

/* from menus.c */
BCINT DisplayMenuArray( BCINT , BCINT , char *, BCINT , BCINT  *, char *[ 30], BCINT  [30]);
BCINT DisplayMenu( BCINT , BCINT , char *, ...);
BCINT PullDownMenu( BCINT , BCINT , ...);
BCINT InputInteger( BCINT , BCINT , BCINT  *, BCINT , BCINT );
Bool Input2Numbers( BCINT , BCINT , char *, char *, BCINT,  BCINT, BCINT * , BCINT * );
BCINT InputIntegerValue( BCINT , BCINT , BCINT , BCINT , BCINT );
BCINT InputIntegerValuePlain( BCINT , BCINT , BCINT , BCINT , BCINT );
void InputNameFromListWithFunc( BCINT , BCINT , char *, BCINT , char **, BCINT , char *, BCINT , BCINT , void (*hookfunc)(BCINT , BCINT , BCINT , BCINT , char *));
BCINT BrowseWalls(char *);
BCINT BrowseFlats(char *);
	 
void InputNameFromList( BCINT , BCINT , char *, BCINT , char **, char *);
void InputFileName( BCINT , BCINT , char *, BCINT , char *);
void InputCommand( BCINT , BCINT , char *, BCINT , char *);
Bool Confirm( BCINT , BCINT , char *, char *);
void Notify( BCINT , BCINT , char *, char *);
void DisplayMessage( BCINT , BCINT , char *, ...);
void NotImplemented( void);

/* from objects.c */
BCINT GetMaxObjectNum(BCINT );
void HighlightSelection( BCINT , SelPtr);
Bool IsSelected( SelPtr, BCINT );
SelPtr SelectPointer( SelPtr list, BCINT objnum );
SelPtr SelectObject1( SelPtr *, BCINT );
SelPtr SelectObjectAfter( SelPtr, BCINT );
void UnSelectObject( SelPtr *, BCINT );
void ForgetSelection( SelPtr *);
BCINT GetMaxObjectNum( BCINT);
BCINT GetCurObject( int , double , double , double );
Bool CurObjectUnderMouse( BCINT , BCINT , BCINT , BCINT , BCINT, BCINT ); 
SelPtr SelectObjectsInBox( BCINT , BCINT , BCINT , BCINT , BCINT );
void HighlightObject( BCINT , BCINT , BCINT );
void DeleteObject1( BCINT , BCINT );
void DeleteObjects( BCINT , SelPtr *);
void InsertObject( BCINT , BCINT , BCINT , BCINT );
BCINT GetOppositeSector( BCINT, Bool); /* SWAP ! */
Bool IsLineDefInside( BCINT , BCINT , BCINT , BCINT , BCINT ); /* SWAP - needs Vertexes & LineDefs */
void CopyObjects( BCINT , SelPtr);
Bool MoveObjectsToCoords( BCINT , SelPtr, BCINT , BCINT , BCINT );
void GetObjectCoords( BCINT , BCINT , BCINT  *, BCINT  *);
void RotateAndScaleObjects( BCINT , SelPtr, double, double, int);
BCINT FindFreeTag(void);
int FindFreeLineID(void);
int FindFreePolyID(void);
void FlipLineDefs( SelPtr, Bool);
void DeleteVerticesJoinLineDefs( SelPtr );
void MergeVertices( SelPtr *);
Bool AutoMergeVertices( SelPtr *);
void SplitLineDefs( SelPtr);
void SplitSector( BCINT , BCINT );
void SplitLineDefsAndSector( BCINT , BCINT );
void ConnectLineDefsSplitDonut( BCINT linedef1, BCINT linedef2); 
void ConnectLineDefsMakeHallway( BCINT linedef1, BCINT linedef2);
void MergeSectors( SelPtr *);
void DeleteLineDefsJoinSectors( SelPtr *);
void MakeDoorFromSector( BCINT );
void MakeLiftFromSector( BCINT );
void AlignTexturesY( SelPtr *, Bool, Bool);
void AlignTexturesX( SelPtr *, Bool, Bool);
void DistributeSectorFloors( SelPtr);
BCINT CommonVertex(BCINT, BCINT);
SelPtr rev_list(SelPtr);	/* return a new list, which is the reverse
							   of the original */
void delete_list(SelPtr);	/* use this to delete it afterwards */
void DistributeSectorCeilings( SelPtr);
void DistributeLightLevels( SelPtr);


/* from editobj.c */
void DisplayObjectInfo( BCINT , BCINT );
BCINT InputObjectNumber( BCINT , BCINT , BCINT , BCINT );
BCINT InputObjectXRef( BCINT , BCINT , BCINT , Bool, BCINT );
Bool Input2VertexNumbers( BCINT, BCINT, char *, BCINT *, BCINT *);
Bool Input2NegativeNumbers( BCINT, BCINT, char *, char *, BCINT, BCINT, BCINT *, BCINT *);
void EditObjectsInfo( BCINT , BCINT , BCINT , SelPtr);
BCINT CheckLevel( BCINT , BCINT ); 
void CheckStartingPos( void); 
Bool CheckThingHeights( void);
BCINT InsertStandardObject( BCINT , BCINT , BCINT , BCINT );
BCINT MiscOperations( BCINT , BCINT , BCINT , SelPtr *, BCINT);
void Preferences( BCINT , BCINT );
BCINT SelectThingType(void);
thing_class *SelectThingClass(void);
scheme_class *SelectSchemeClass(void);
Bool Failed;
Bool CheckReport; //jff
char *GetLineDefFlagsLongName(BCINT);
BCINT SelectSpawnType(void);
BCINT SelectNegative(void);
BCINT SelectHexenAngle(void);
spawn_class *SelectSpawnClass(void);
BCINT SelectEnumValue(char *enumName);

/* from draw.c */
extern Bool GridShown;
extern Bool FloorGridShown;
extern Bool GridAccurate;
extern Bool DrawThingCircles;

void CenterMapAroundCoords( BCINT , BCINT );
void ShowAllMap (void);

/* from textures.c */
void ChooseFloorTexture( BCINT , BCINT , char *, BCINT , char **, char *);
void ChooseWallTexture( BCINT , BCINT , char *, BCINT , char **, char *);
void ChooseSprite( BCINT , BCINT , char *, char *);
void GetWallTextureSize( BCINT *, BCINT *, char *);
void *GetResource(char *);
void ForgetResource(char *);
Texture *FindTexture(char *);
Texture *FindTextureIn(char *, void *);
void ForgetAllResources(void);
void GetPicSize(char *, BCINT *, BCINT *);
void DisplayFloorTexture(BCINT, BCINT, BCINT, BCINT, char *);
void DisplayPic(BCINT, BCINT, BCINT, BCINT, char *, Bool);
void DisplayPicFit(BCINT, BCINT, BCINT, BCINT, char *, Bool);
void DisplaySprite(BCINT, BCINT, BCINT, BCINT, char *);
void DisplayWallQuick(BCINT, BCINT, BCINT, BCINT, char *, BCINT);
void DisplayPicFrac( BCINT x0, BCINT y0, BCINT x1, BCINT y1, char *picname, long xstep, long ystep);

/* from readcfg.c */
void readcfg(char *);
void fixcfg(char *);
SList SList_append(SList, char *);
BCINT SelectScheme(void);


/* from mainloop.c */
extern BCINT EditMode;

char  *GetWadFileName();
void  SaveConfigFileOptions();
void  DrawMapQuick(void);
void  Options(BCINT, BCINT);
void  SetMouseColors(BCINT, BCINT);
void  DrawInfoBar(BCINT, BCINT);

Bool  MouseInArea(BCINT, BCINT, BCINT, BCINT);

BCINT FunctionSectorLightLevel(BCINT, SelPtr);
BCINT FunctionSectorCeilingHeight(BCINT, SelPtr);
BCINT FunctionSectorFloorHeight(BCINT, SelPtr);
BCINT FunctionSectorType(BCINT, SelPtr);
BCINT FunctionSectorFlags (BCINT, SelPtr);
BCINT FunctionSectorBrowseFloor(BCINT, SelPtr);
BCINT FunctionSectorBrowseCeiling(BCINT, SelPtr);
BCINT FunctionSectorRef(BCINT, SelPtr);
BCINT FunctionSectorTag(BCINT, SelPtr);
BCINT FunctionRenumberSectorsLowest(BCINT CurObject, SelPtr Selected);

BCINT FunctionVertexSplitSector(BCINT, SelPtr);
BCINT FunctionVertexMerge(BCINT, SelPtr *);
BCINT FunctionVertexSplitLineDef(BCINT, SelPtr);
BCINT FunctionVertexDeleteJoin(BCINT, SelPtr);

BCINT FunctionThingFlag(BCINT, SelPtr, BCINT);
BCINT FunctionThingFlags(BCINT, SelPtr);
BCINT FunctionThingType(BCINT, SelPtr);
BCINT FunctionThingSpecial(BCINT, SelPtr);
BCINT FunctionThingAngle(BCINT, SelPtr);
BCINT FunctionThingXPos(BCINT, SelPtr);
BCINT FunctionThingYPos(BCINT, SelPtr);
BCINT FunctionThingDeathScript(BCINT, SelPtr);

BCINT FunctionThingZPos(BCINT, SelPtr);
BCINT FunctionThingTID(BCINT, SelPtr);
BCINT FunctionThingHexenAngle(BCINT, SelPtr, BCINT);
BCINT FunctionThingNegative(BCINT, SelPtr, BCINT);
BCINT FunctionThingSpawnType(BCINT, SelPtr, BCINT);
BCINT FunctionThingArg(BCINT, SelPtr, BCINT);
BCINT FunctionThingEnum(BCINT, SelPtr, BCINT, char *enumName);
BCINT FunctionThingArgFlags(BCINT, SelPtr, BCINT, char *flagsName);

BCINT FunctionLineDefFlag(BCINT, SelPtr, BCINT);
BCINT FunctionLineDefFlags(BCINT, SelPtr);
BCINT FunctionLineDefArgFlags(BCINT, SelPtr, int);
BCINT FunctionLineDefSplitAdd(BCINT, SelPtr);
BCINT FunctionConnectLineDefsSplitDonut(BCINT CurObject, SelPtr Selected);
BCINT FunctionConnectLineDefsMakeHallway(BCINT CurObject, SelPtr Selected);
BCINT FunctionRenumberLineDefsLowest(BCINT CurObject, SelPtr Selected);
BCINT FunctionLineDefSwapSideDefs(BCINT, SelPtr);
BCINT FunctionLineDefSpecial(BCINT, SelPtr);
BCINT FunctionLineDefDeleteJoin(BCINT, SelPtr);
BCINT FunctionLineDefFlipSideDefs(BCINT, SelPtr);
BCINT FunctionLineDefBrosweWallTextures(BCINT, SelPtr);
BCINT FunctionLineDefTag(BCINT, SelPtr);
BCINT FunctionLineDefWallTexture(BCINT, SelPtr, char *, BCINT, BCINT);
BCINT FunctionLineDefWallTextures(BCINT, SelPtr, BCINT, BCINT);
BCINT FunctionLineDefBrowseWallTextures(BCINT, SelPtr, BCINT, BCINT);
BCINT FunctionLineDefXoffset(BCINT, SelPtr, BCINT);
BCINT FunctionLineDefYoffset(BCINT, SelPtr, BCINT);
BCINT FunctionLineDefSectorRef(BCINT, SelPtr, BCINT);

BCINT FunctionLineDefTrigger(BCINT, SelPtr);
BCINT FunctionLineDefHexenAngle(BCINT, SelPtr, BCINT);
BCINT FunctionLineDefNegative(BCINT, SelPtr, BCINT);
BCINT FunctionLineDefSpawnType(BCINT, SelPtr, BCINT);
BCINT FunctionLineDefArg(BCINT, SelPtr, BCINT);
BCINT FunctionLineDefEnum(BCINT, SelPtr, BCINT);

BCINT OnScreenEditingSideDefTextures(BCINT, SelPtr);
BCINT OnScreenEditingSideDef1(BCINT, SelPtr);
BCINT OnScreenEditingSideDef2(BCINT, SelPtr);
BCINT OnScreenEditingSectors(BCINT, SelPtr);
BCINT OnScreenEditingSectorFlats(BCINT, SelPtr);
BCINT OnScreenEditingLineDefs(BCINT, SelPtr);
BCINT OnScreenEditingThings(BCINT, SelPtr);

char *BroweFlats(BCINT);
void DisplayAbout();
void DisplayHelp(BCINT, BCINT);
void HelpThingBox();
void HelpSectorBox();
void HelpTextureBox();
void DisplayHotKeys(BCINT);

#if defined(_MSC_VER)

#pragma warning (disable: 4035)
static __inline long MulScale19 (long a, long b)
{
	__asm mov eax,a
	__asm imul b
	__asm shrd eax,edx,19
}

static __inline long DivScale19 (long a, long b)
{
	__asm mov edx,a
	__asm sar edx,13
	__asm mov eax,a
	__asm shl eax,19
	__asm idiv b
}
#pragma warning (default: 4035)

#elif defined(__WATCOMC__)

#pragma aux MulScale19 =\
	"imul edx",\
	"shrd eax, edx, 19",\
	parm nomemory [eax][edx]\
	modify exact [eax edx]\

#pragma aux DivScale19 =\
	"mov edx, eax",\
	"sar edx, 13",\
	"shl eax, 19",\
	"idiv ebx",\
	parm nomemory [eax][ebx]\
	modify exact [eax edx]\

#else

static __inline__ long MulScale19 (long a, long b)
{
	long result, dummy;

	asm volatile
		("imull %3\n\tshrdl %b4,%1,%0"
		: "=a,a" (result), "=d,d" (dummy)
		: "a,a" (a), "m,r" (b), "I,I" (19)
		: "%cc"
	);
	return result;
}

static __inline__ long DivScale19 (long a, long b)
{
        long result, dummy;

        asm volatile
                ("idivl %4"
                :"=a,a" (result),
                 "=d,d" (dummy)
                : "a,a" (a<<19),
                  "d,d" (a>>13),
                  "r,m" (b)
                : "%cc");
        return result;
}
#endif

/* end of file */
