/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.
   
   GFX.C - Graphics routines.
   */

/* the includes */
#include "deu.h"
#include <math.h>
#include <dos.h>

#define BCabs(x)		((x < 0) ? -(x) : (x))

BYTE DPal[768];
BCINT Palette[PALETTE_SIZE];
/*
void ScreenBoxArea(int, int, int, int, int);
*/

extern GrFont DethFont;
extern GrFont SmallFont;


#define BGI_PATH "."

/* the global variables */
BCINT GfxMode = 0;		/* graphics mode number, or 0 for text */
/* 1 = 320x200, 2 = 640x480, 3 = 800x600, 4 = 1024x768, 5 = 1280x960, 6 = 1600x1200 */
/* positive = 16 colors, negative = 256 colors */
double OrigX;			/* the X origin */
double OrigY;			/* the Y origin */
double Scale;			/* the scale value */
BCINT PointerX;			/* X position of pointer */
BCINT PointerY;			/* Y position of pointer */
int ScrMaxX;			/* maximum X screen coord */
int ScrMaxY;			/* maximum Y screen coord */
int ScrCenterX;			/* X coord of screen center */
int ScrCenterY;			/* Y coord of screen center */
GrTextOption TextOption;
GrLineOption LineOption;
Bool CustomLine;

#ifndef _WIN32
static int res[7][3] = {{640,480,16},{320,200,256},{640,480,256},
						{800,600,256},{1024,768,256},{1280,960,256},{1600,1200,256}};
#endif


/*
   initialise the graphics display
   */

void InitGfx()
{
    static int  gdriver;
    static int  gmode;
    int         errorcode = grNoInitGraph;

    FakeCursor = FALSE;
    
    if (VideoMode < 0 || VideoMode > 6)
		VideoMode = 0;
    set_BGI_mode_whc( &gdriver, &gmode, res[VideoMode][0], res[VideoMode][1], res[VideoMode][2]);
    initgraph( &gdriver, &gmode, BGI_PATH);
  //  registerbgifont(&_litt_font);
    errorcode = graphresult();
    if(errorcode != grOk)
		ProgError( "graphics error: %s", grapherrormsg( errorcode));
    
    if (VideoMode == 0)
		GfxMode = 2; /* 640x480x16 */
    else
	{
		GfxMode = -VideoMode; /* 640x480x256, 800x600x256, or 1024x768x256 */
		SetDoomPalette( 0);
    }
    
    SetLineDotted (FALSE);
    setbkcolor( TranslateToGameColor( BLACK));
    ScrMaxX = getmaxx();
    ScrMaxY = getmaxy();
    ScrCenterX = ScrMaxX / 2;
    ScrCenterY = ScrMaxY / 2;

	TextOption.txo_font = &DethFont;
	TextOption.txo_fgcolor.v = 0;
	TextOption.txo_bgcolor.v = GrNOCOLOR;
	TextOption.txo_chrtype = GR_BYTE_TEXT;
	TextOption.txo_direct = GR_TEXT_RIGHT;
	TextOption.txo_xalign = GR_ALIGN_LEFT;
	TextOption.txo_yalign = GR_ALIGN_TOP;

	LineOption.lno_color = 0;
	LineOption.lno_width = 1;
	LineOption.lno_pattlen = 0;
	LineOption.lno_dashpat = "\0\2\2\0";
}



/*
   terminate the graphics display
   */

void TermGfx()
{
	union REGS regs;

    if (GfxMode)
	{
		GfxMode = 0;
		// Let Windows know that we really are in a text mode
		memset (&regs, 0, sizeof(regs));
		regs.w.ax = 3;
		int386(0x10, &regs, &regs);
		// Let the graphics routines know we are in a text mode
		closegraph();
    }
}


/*
   switch from VGA 16 colours to VGA 256 colours
   */

Bool SwitchToVGA256()
{
    static int gdriver = -1;
    int gmode, errorcode;
    
    if (GfxMode > 0) {
		closegraph();
		set_BGI_mode_whc( &gdriver, &gmode, res[1][0], res[1][1], res[1][2]);
		initgraph( &gdriver, &gmode, BGI_PATH);
		errorcode = graphresult();
		if(errorcode != grOk) {
			/* failed for 256 colors - back to 16 colors */
			set_BGI_mode_whc( &gdriver, &gmode, res[0][0], res[0][1], res[0][2]);
			initgraph( &gdriver, &gmode, BGI_PATH);
			errorcode = graphresult();
		}
		
		
		if (errorcode != grOk) /* shouldn't happen */
			ProgError( "graphics error: %s", grapherrormsg( errorcode));
		GfxMode = -1 /* 320x200x256 */;
		SetDoomPalette( 0);
		ScrMaxX = getmaxx();
		ScrMaxY = getmaxy();
		ScrCenterX = ScrMaxX / 2;
		ScrCenterY = ScrMaxY / 2;
		return TRUE;
	}

    return FALSE;
}


/*
   switch from VGA 256 colours to VGA 16 colours
   */

Bool SwitchToVGA16()
{
    if (GfxMode == -1) { 
		TermGfx();   
		InitGfx();  
		CheckMouseDriver();
		if (UseMouse)
			ShowMousePointer();
		return TRUE;
    }
    return FALSE;
}



/*
   clear the screen
   */

void ClearScreen()
{
    cleardevice();
}

/*
   clear the map screen
   */




/*AJB*/
void ClearMapScreen( BCINT top, BCINT maxy)
{
	GrSetClipBox (0, top, ScrMaxX, ScrMaxY - maxy);
	GrClearClipBox (0);
	GrResetClipBox ();
}



/*
   set the current drawing color
   */

void SetColor( int color)
{
    if (GfxMode < 0)
		color = TranslateToGameColor(color);
	setcolor( color);
	LineOption.lno_color = (LineOption.lno_color & GrCMODEMASK) | color;
	TextOption.txo_fgcolor.v = LineOption.lno_color;
}

/*
	Set the text background color (-1 is none)
	*/

void SetTextBg( int color)
{
	if (color < 0)
	{
		TextOption.txo_bgcolor.v = GrNOCOLOR;
	}
	else
	{
		color = TranslateToGameColor(color);
		TextOption.txo_bgcolor.v = color | GrWRITE;
	}
}

/*
	Enable or disable horizontal text centering
	*/
void SetTextCentered (Bool center)
{
	TextOption.txo_xalign = center ? GR_ALIGN_CENTER : GR_ALIGN_LEFT;
}

/*
	set the write mode
	*/

void SetWriteMode (int mode)
{
	setwritemode (mode);
	switch (mode)
	{
	default:
	case COPY_PUT:		mode = GrWRITE;		break;
	case XOR_PUT:		mode = GrXOR;		break;
	case OR_PUT:		mode = GrOR;		break;
	case AND_PUT:		mode = GrAND;		break;
	}
	LineOption.lno_color = (LineOption.lno_color & GrCVALUEMASK) | mode;
	TextOption.txo_fgcolor.v = LineOption.lno_color;
}

void SetLineDotted (Bool dotted)
{
	LineOption.lno_pattlen = dotted ? 4 : 0;
	CustomLine = dotted || LineOption.lno_width > 1;
}

void SetLineThickness (int width)
{
	LineOption.lno_width = width;
	CustomLine = width > 1 || LineOption.lno_pattlen != 0;
}

/*
   draw a line on the screen from map coords
   */

void DrawMapLine( int mapXstart, int mapYstart, int mapXend, int mapYend)
{
	if ((SCREENX(mapXstart) < -20000) || (SCREENX(mapXstart) > 20000) || (SCREENY(mapYstart) < -20000) || (SCREENY(mapYstart) > 20000))
		return;
	DrawScreenLine(SCREENX( mapXstart), SCREENY( mapYstart), SCREENX( mapXend), SCREENY( mapYend));
}



/*
   draw a circle on the screen from map coords
   */

void DrawMapCircle( BCINT mapXcenter, BCINT mapYcenter, BCINT mapRadius)
{
	if ((SCREENX(mapXcenter) < -20000) || (SCREENX(mapXcenter) > 20000) || (SCREENY(mapYcenter) < -20000) || (SCREENY(mapYcenter) > 20000))
		return;
	circle( SCREENX( mapXcenter), SCREENY( mapYcenter), (int) (mapRadius * Scale));
}



/*
   draw an arrow on the screen from map coords
   */

void DrawMapVector( BCINT mapXstart, BCINT mapYstart, BCINT mapXend, BCINT mapYend)
{
	BCINT  scrXstart = SCREENX( mapXstart);
	BCINT  scrYstart = SCREENY( mapYstart);
	BCINT  scrXend   = SCREENX( mapXend);
	BCINT  scrYend   = SCREENY( mapYend);
	double r         = hypot( (double) (scrXstart - scrXend), (double) (scrYstart - scrYend));
	BCINT  scrXoff   = (r >= 1.0) ? (BCINT) ((scrXstart - scrXend) * 4.0 / r * Scale) : 0;
	BCINT  scrYoff   = (r >= 1.0) ? (BCINT) ((scrYstart - scrYend) * 4.0 / r * Scale) : 0;
	
	if ((SCREENX(mapXstart) < -20000) || (SCREENX(mapXstart) > 20000) || (SCREENY(mapYstart) < -20000) || (SCREENY(mapYstart) > 20000))
		return;
	DrawScreenLine( scrXstart, scrYstart, scrXend, scrYend);

	scrXstart = scrXend + 2 * scrXoff;
	scrYstart = scrYend + 2 * scrYoff;
	DrawScreenLine( scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend);
	DrawScreenLine( scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend);
}



/*
   draw an arrow on the screen from map coords and angle (0 - 65535)
   */

void DrawMapArrow( BCINT mapXstart, BCINT mapYstart, UBCINT angle)
{
	BCINT  mapXend   = mapXstart + (BCINT) (50 * cos(angle / 10430.37835));
	BCINT  mapYend   = mapYstart + (BCINT) (50 * sin(angle / 10430.37835));
	BCINT  scrXstart = SCREENX( mapXstart);
	BCINT  scrYstart = SCREENY( mapYstart);
	BCINT  scrXend   = SCREENX( mapXend);
	BCINT  scrYend   = SCREENY( mapYend);
	double r         = hypot( scrXstart - scrXend, scrYstart - scrYend);
	BCINT  scrXoff   = (r >= 1.0) ? (BCINT) ((scrXstart - scrXend) * 4.0 / r * Scale) : 0;
	BCINT  scrYoff   = (r >= 1.0) ? (BCINT) ((scrYstart - scrYend) * 4.0 / r * Scale) : 0;

	if ((SCREENX(mapXstart) < -20000) || (SCREENX(mapXstart) > 20000) || (SCREENY(mapYstart) < -20000) || (SCREENY(mapYstart) > 20000))
		return;
	
	DrawScreenLine( scrXstart, scrYstart, scrXend, scrYend);
	scrXstart = scrXend + 2 * scrXoff;
	scrYstart = scrYend + 2 * scrYoff;
	DrawScreenLine( scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend);
	DrawScreenLine( scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend);
}

/*
	draw a filled box on the map
	*/

void DrawMapBox (int x1, int y1, int x2, int y2)
{
	x1 = SCREENX (x1);
	x2 = SCREENX (x2);
	y1 = SCREENY (y1);
	y2 = SCREENY (y2);

	if (abs (x1) > 20000 || abs (y1) > 20000)
		return;

	GrFilledBox (x1, y1, x2, y2, LineOption.lno_color);
}

/*
	draw a filled box on the map with a black frame.
	*/
void DrawMapTBox (int x1, int y1, int x2, int y2)
{
	x1 = SCREENX (x1);
	x2 = SCREENX (x2);
	y1 = SCREENY (y1);
	y2 = SCREENY (y2);

	if (abs (x1) <= 20000 || abs (y1) <= 20000)
	{
		GrFilledBox (x1, y1, x2, y2, LineOption.lno_color);
		if (abs(x2-x1) >= 5 && abs(y2-y1) >= 5)
			GrBox (x1, y1, x2, y2, TranslateToGameColor(BLACK) | GrWRITE);
	}
}

/*
	draw a hollow box on the map
	*/

void DrawMapHollowBox (int x1, int y1, int x2, int y2)
{
	x1 = SCREENX (x1);
	x2 = SCREENX (x2);
	y1 = SCREENY (y1);
	y2 = SCREENY (y2);

	if (abs (x1) > 20000 || abs (y1) > 20000)
		return;

	if (!CustomLine)
	{
		GrBox (x1, y1, x2, y2, LineOption.lno_color);
	}
	else
	{
		GrCustomBox (x1, y1, x2, y2, &LineOption);
	}
}

/*
   draw a line on the screen from screen coords
   */

void DrawScreenLine( BCINT Xstart, BCINT Ystart, BCINT Xend, BCINT Yend)
{
	if (!CustomLine)
	{
		GrLine (Xstart, Ystart, Xend, Yend, LineOption.lno_color);
	}
	else
	{
		GrCustomLine (Xstart, Ystart, Xend, Yend, &LineOption);
	}
}

/*
   draw a filled in box on the screen from screen coords
   */

void DrawScreenBox( BCINT Xstart, BCINT Ystart, BCINT Xend, BCINT Yend)
{
	GrFilledBox (Xstart, Ystart, Xend, Yend, TextOption.txo_fgcolor.v);
}

void DrawScreenFrame( BCINT Xstart, BCINT Ystart, BCINT Xend, BCINT Yend)
{
	GrBox (Xstart, Ystart, Xend, Yend, TextOption.txo_fgcolor.v);
}


/*
   draw a filled-in 3D-box on the screen from screen coords
   */

void DrawScreenBox3D( BCINT Xstart, BCINT Ystart, BCINT Xend, BCINT Yend)
{
    setfillstyle( 1, TranslateToGameColor(DARKGRAY));
    bar( Xstart + 1, Ystart + 1, Xend - 1, Yend - 1);
    SetColor(DARKERGRAY);
    line( Xstart, Yend, Xend, Yend);
    line( Xend, Ystart, Xend, Yend);
    SetColor(LIGHTGRAY);
    line( Xstart, Ystart, Xend, Ystart);
    line( Xstart, Ystart, Xstart, Yend);
    SetColor( BLACK);
}



/*
   draw a hollow 3D-box on the screen from screen coords
   */

void DrawScreenBoxHollow( BCINT Xstart, BCINT Ystart, BCINT Xend, BCINT Yend)
{
    setfillstyle( 1, TranslateToGameColor( BLACK));
    bar( Xstart + 1, Ystart + 1, Xend - 1, Yend - 1);
    SetColor(LIGHTGRAY);
    line( Xstart, Yend, Xend, Yend);
    line( Xend, Ystart, Xend, Yend);
    SetColor(DARKERGRAY);
    line( Xstart, Ystart, Xend, Ystart);
    line( Xstart, Ystart, Xstart, Yend);
    SetColor(LIGHTGRAY);
}



void DrawScreenBar( BCINT Xstart, BCINT Ystart, BCINT Width, Bool Reversed)
{
	SetColor(Reversed ? DARKERGRAY : LIGHTGRAY);
    line( Xstart, Ystart, Xstart + Width, Ystart);
	SetColor(Reversed ? LIGHTGRAY : DARKERGRAY);
    line( Xstart, Ystart + 1, Xstart + Width, Ystart + 1);
}

// [RH] Why did the buttons have an illegible bright shadow?
void DrawScreenButton( BCINT Xstart, BCINT Ystart, BCINT Width, BCINT Height, char *Text, Bool Filled)
{
	if (Filled == TRUE) {
		setfillstyle( 1, TranslateToGameColor( DARKGRAY));
		bar( Xstart + 1, Ystart + 1, Xstart + Width - 1, Ystart + Height - 1);
   	}
    SetColor(DARKERGRAY);
    line( Xstart, Ystart + Height, Xstart + Width, Ystart + Height);
    line( Xstart + Width, Ystart, Xstart + Width, Ystart + Height); 
	SetColor(LIGHTGRAY);
	line( Xstart, Ystart, Xstart + Width, Ystart);
	line( Xstart, Ystart, Xstart, Ystart + Height); 
	ScreenBoxArea(Xstart - 1, Ystart - 1, Width + 2, Height + 2, BLACK);
	if (Text != "" || Text != NULL)
	{
		TextOption.txo_font = &SmallFont;
		TextOption.txo_xalign = GR_ALIGN_CENTER;
		SetColor (BLACK);
		GrDrawString (Text, strlen (Text), Xstart + (Width / 2), Ystart + (Height / 2) - 6, &TextOption);
		TextOption.txo_font = &DethFont;
		TextOption.txo_xalign = GR_ALIGN_LEFT;
	}
}

void DrawScreenButtonGray( BCINT Xstart, BCINT Ystart, BCINT Width, BCINT Height, char *Text, Bool Filled)
{
	if (Filled == TRUE)
	{
		setfillstyle( 1, TranslateToGameColor( DARKGRAY));
		bar( Xstart + 1, Ystart + 1, Xstart + Width - 1, Ystart + Height - 1);
   	}
    SetColor(DARKERGRAY);
    line( Xstart, Ystart + Height, Xstart + Width, Ystart + Height);
    line( Xstart + Width, Ystart, Xstart + Width, Ystart + Height); 
	SetColor(LIGHTGRAY);
	line( Xstart, Ystart, Xstart + Width, Ystart);
	line( Xstart, Ystart, Xstart, Ystart + Height); 
	ScreenBoxArea(Xstart - 1, Ystart - 1, Width + 2, Height + 2, BLACK);
	if (Text != "" || Text != NULL)
	{
		TextOption.txo_font = &SmallFont;
		TextOption.txo_xalign = GR_ALIGN_CENTER;
		SetColor (DARKERGRAY);
		GrDrawString (Text, strlen (Text), Xstart + (Width / 2), Ystart + (Height / 2) - 6, &TextOption);
		TextOption.txo_font = &DethFont;
		TextOption.txo_xalign = GR_ALIGN_LEFT;
	}
}

void DrawScreenButtonIn( BCINT Xstart, BCINT Ystart, BCINT Width, BCINT Height, char *Text, Bool Filled)
{
	if (Filled)
	{
		setfillstyle( 1, TranslateToGameColor( DARKGRAY));
		bar( Xstart + 1, Ystart + 1, Xstart + Width - 1, Ystart + Height - 1);
   	}
	SetColor(LIGHTGRAY);
	line( Xstart, Ystart + Height, Xstart + Width, Ystart + Height);
	line( Xstart + Width, Ystart, Xstart + Width, Ystart + Height); 
	SetColor(DARKERGRAY);
	line( Xstart, Ystart, Xstart + Width, Ystart);
	line( Xstart, Ystart, Xstart, Ystart + Height); 
	if (Text != " " || Text != NULL)
	{
		TextOption.txo_font = &SmallFont;
		TextOption.txo_xalign = GR_ALIGN_CENTER;
		SetColor (BLACK);
		GrDrawString (Text, strlen (Text), Xstart + (Width / 2), Ystart + (Height / 2) - 6, &TextOption);
		TextOption.txo_font = &DethFont;
		TextOption.txo_xalign = GR_ALIGN_LEFT;
	}
}

void DrawScreenButtonOut( BCINT Xstart, BCINT Ystart, BCINT Width, BCINT Height, Bool Filled)
{
	if (Filled)
	{
		setfillstyle( 1, TranslateToGameColor( BLUE));
		bar( Xstart + 1, Ystart + 1, Xstart + Width - 1, Ystart + Height - 1);
   	}
	SetColor(DARKERGRAY);
	line( Xstart, Ystart + Height, Xstart + Width, Ystart + Height);
	line( Xstart + Width, Ystart, Xstart + Width, Ystart + Height); 
	SetColor(LIGHTGRAY);
	line( Xstart, Ystart, Xstart + Width, Ystart);
	line( Xstart, Ystart, Xstart, Ystart + Height); 
}


void DrawScreenWindow(int x, int y, int width, int height, char *title, int color)
{
	int dx = x + width;
	int dy = y + height;
	
	SetColor(LIGHTGRAY);
	line(x, y, dx - 1, y);
	line(x, y, x, dy - 1);
	SetColor(DARKERGRAY);
	line(x, dy, dx , dy);
	line(dx, y, dx, dy);
	setfillstyle( 1, TranslateToGameColor(DARKGRAY));
	bar( x + 1, y + 1, dx - 1, dy - 1);
	setfillstyle( 1, TranslateToGameColor(color));
	bar( x + 3, y + 3, dx - 3, y + 17);
	SetColor(BLACK);
	DrawScreenTextSmall(x + 6, y + 4, "%s", title);
}


void ScreenHighlightArea(int x, int y, int width, int height, char *text, int color, int textcolor)
{
	int dx = x + width;
	int dy = y + height;

	setfillstyle( 1, TranslateToGameColor(color));
	bar( x + 1, y + 1, dx - 1, dy - 1);
	SetColor(textcolor);
	DrawScreenText(x + 3, y + 2, "%s", text);

}

void ScreenBoxArea(int x, int y, int width, int height, int color)
{
	int dx = x + width;
	int dy = y + height;

	SetColor(color);
    line( x, y, dx, y);
    line( x, y, x, dy);
    line( dx, y, dx, dy);
    line( x, dy, dx, dy);
}

/*
   draw a meter bar on the screen from screen coords (in a hollow box); max. value = 1.0
   */

void DrawScreenMeter( BCINT Xstart, BCINT Ystart, BCINT Xend, BCINT Yend, float value)
{
    if (value < 0.0)
		value = 0.0;
    if (value > 1.0)
		value = 1.0;
    setfillstyle( 1, TranslateToGameColor( BLACK));
    bar( Xstart + 1 + (BCINT) ((Xend - Xstart - 2) * value), Ystart + 1, Xend - 1, Yend - 1);
    setfillstyle( 1, TranslateToGameColor( LIGHTGREEN));
    bar( Xstart + 1, Ystart + 1, Xstart + 1 + (BCINT) ((Xend - Xstart - 2) * value), Yend - 1);
}


/*
   write text to the screen
   */

void DrawScreenText( BCINT Xstart, BCINT Ystart, char *msg, ...)
{
    static BCINT lastX;
    static BCINT lastY;
    char temp[ 120];
	int len;
    va_list args;

    va_start( args, msg);
    len = vsprintf( temp, msg, args);
    va_end( args);

	if (Xstart < -1 || Ystart < -1)
		return;
    if (Xstart == -1)
		Xstart = lastX;
    if (Ystart == -1)
		Ystart = lastY;
	TextOption.txo_font = &DethFont;
	GrDrawString (temp, len, Xstart, Ystart, &TextOption);
    lastX = Xstart;
    lastY = Ystart + 10; 
}

void DrawScreenTextSmall(BCINT Xstart, BCINT Ystart, char *msg, ...)
{
    static BCINT lastX;
    static BCINT lastY;
    char temp[ 120];
	int len;

    va_list args;
    va_start(args, msg);
    len = vsprintf(temp, msg, args);
    va_end( args);

	if (Xstart < -1 || Ystart < -1)
		return;
    if (Xstart == -1)
		Xstart = lastX;
    if (Ystart == -1)
		Ystart = lastY;
	TextOption.txo_font = &SmallFont;
	GrDrawString (temp, len, Xstart, Ystart, &TextOption);
	TextOption.txo_font = &DethFont;
    lastX = Xstart;
    lastY = Ystart + 10;
}

void DrawScreen3DText( BCINT Xstart, BCINT Ystart, BCINT col, BCINT scol, char *msg, ...)
{
    static BCINT lastX;
    static BCINT lastY; 
    char temp[ 120];
	int len;

    va_list args;
    va_start( args, msg);
    len = vsprintf( temp, msg, args);
    va_end( args);

	if (Xstart < -1 || Ystart < -1)
		return;
    if (Xstart == -1)
		Xstart = lastX;
    if (Ystart == -1)
		Ystart = lastY;
	SetColor (scol);
	GrDrawString (temp, len, Xstart, Ystart, &TextOption);
	SetColor (col);
	GrDrawString (temp, len, Xstart - 1, Ystart - 1, &TextOption);

    lastX = Xstart;
    lastY = Ystart + 10; 
}


/*
   draw (or erase) the pointer if we aren't using the mouse
   */

void DrawPointer()
{
	/* use XOR mode : drawing the pointer twice erases it */
	if (FakeCursor)
	{
		SetWriteMode( XOR_PUT);
		SetColor( YELLOW);
		line( PointerX - 12, PointerY - 12, PointerX + 12, PointerY + 12);
		line( PointerX - 12, PointerY + 12, PointerX + 12, PointerY - 12);
		/* restore normal write mode */
		SetWriteMode( COPY_PUT);
	}
}



/*
   load one "playpal" palette and change all palette colours
   */

void SetDoomPalette( BCINT playpalnum)
{
	BYTE	gamma[256];
	MDirPtr	dir;
	BCINT	n;

	if (playpalnum < 0 || playpalnum > 13)
		return;
	dir = FindMasterDir( MasterDir, "PLAYPAL");
	//dir = 0;
	if (dir)
	{
		BasicWadSeek( dir->wadfile, dir->dir.start);
		for (n = 0; n <= playpalnum; n++)
			BasicWadRead( dir->wadfile, DPal, 768L);

		if (Gamma > 0 && Gamma <= 4) 
		{
			static const double gmult[5] = {1.0, 0.75, 0.55, 0.4, 0.3};
			double mult = gmult[Gamma];

			for (n = 0; n < 256; ++n)
			{
				gamma[n] = pow (((double)n / 255.0), mult) * 255.0;
			}
		}
		else
		{
			for (n = 0; n < 256; ++n)
			{
				gamma[n] = n;
			}
		}

		GrResetColors();
		for (n = 0; n < 254; n++)
		{
			GrAllocCell();
		}

		for (n = 0; n < 256; n++)
		{
			GrSetColor (n, gamma[DPal[3*n]], gamma[DPal[3*n+1]], gamma[DPal[3*n+2]]);
		}
	}
}

/*
   translate a standard color to Doom palette 0 (approx.)
   */

BCINT TranslateToGameColor( BCINT color)
{
	return Palette[color];
}



/*
   translate (dx, dy) into an integer angle value (0-65535)
   */

UBCINT ComputeAngle( BCINT dx, BCINT dy)
{
    return (UBCINT) (atan2( (double) dy, (double) dx) * 10430.37835 + 0.5);
    /* Yes, I know this function could be in another file, but */
    /* this is the only source file that includes <math.h>...  */
}



/*
   compute the distance from (0, 0) to (dx, dy)
   */

UBCINT ComputeDist( BCINT dx, BCINT dy)
{
    return (UBCINT) (hypot( (double) dx, (double) dy) + 0.5);
    /* Yes, I know this function could be in another file, but */
    /* this is the only source file that includes <math.h>...  */
}



/*
   insert the vertices of a new polygon
   */

void InsertPolygonVertices( BCINT centerx, BCINT centery, BCINT sides, BCINT radius)
{
    BCINT n;
    
    for (n = 0; n < sides; n++)
	{
		InsertObject( OBJ_VERTEXES, -1,
			centerx + (BCINT) ((double) radius * cos( 2*M_PI * (double) n / (double) sides)),
			centery + (BCINT) ((double) radius * sin( 2*M_PI * (double) n / (double) sides)));
	}
    /* Yes, I know... etc. */
}



/*
   move (x, y) to a new position: rotate and scale around (0, 0)
   */

//void RotateAndScaleCoords(BCINT *x, BCINT *y, double angle, double scale)
//{
//    double r, theta;
//    
//    r = hypot((double) *x, (double) *y);
//    theta = atan2((double) *y, (double) *x);
//    *x = (BCINT) (r * scale * cos(theta + angle) + 0.5);
//    *y = (BCINT) (r * scale * sin(theta + angle) + 0.5);
//}
// jff
// use double throughout, round final answer for increased accuracy
//
void RotateAndScaleCoords(double *x, double *y, double angle, double scale)
{
    double r, theta;
    
    r = hypot(*x,*y);
    theta = atan2(*y,*x);
    *x = r * scale * cos(theta + angle);
    *y = r * scale * sin(theta + angle);
}

/* end of file */
